unit CodeGen;

interface

uses
  SysUtils, Parser, Scaner, Linker, Globals;

  procedure AssembleAndLink;
  
implementation

{      }

var
  OCTab : array [1..MaxBinCode] of char;
  CodeAddress   : integer;

procedure OCC(C:char);
begin
  if CodeAddress<MaxBinCode then begin
    CodeAddress:=CodeAddress+1;
    OCTab[CodeAddress]:=C;
  end else begin
    Error(902);
  end;
end;

procedure OC(B:integer);
begin
  OCC(CHR(B));
end;

procedure OCW(I:integer);
begin
  if I>=0 then begin
    OC(I mod 256);
    OC((I div 256) mod 256);
  end else begin
    I:=-(I+1);
    OC(255-(I mod 256));
    OC(255-((I div 256) mod 256));
  end;
end;

procedure OCI(I:integer);
begin
  if I>=0 then begin
    OC(I mod 256);
    OC((I div 256) mod 256);
    OC((I div 65536) mod 256);
    OC(I div 16777216);
  end else begin
    I:=-(I+1);
    OC(255-(I mod 256));
    OC(255-((I div 256) mod 256));
    OC(255-((I div 65536) mod 256));
    OC(255-(I div 16777216));
  end;
end;

function OCGI(O:integer):integer;
begin
  if ord(OCTab[O+3])<$80 then begin
    OCGI:=ord(OCTab[O])+(ord(OCTab[O+1])*256)+(ord(OCTab[O+2])*65536)+(ord(OCTab[O+3])*16777216);
  end else begin
    OCGI:=-(((255-ord(OCTab[O]))+((255-ord(OCTab[O+1]))*256)+((255-ord(OCTab[O+2]))*65536)+((255-ord(OCTab[O+3]))*16777216))+1);
  end;
end;

procedure OCPI(O,I:integer);
begin
  if I>=0 then begin
    OCTab[O]  :=CHR(I mod 256);
    OCTab[O+1]:=CHR((I div 256) mod 256);
    OCTab[O+2]:=CHR((I div 65536) mod 256);
    OCTab[O+3]:=CHR(I div 16777216);
  end else begin
    I:=-(I+1);
    OCTab[O]  :=CHR(255-(I mod 256));
    OCTab[O+1]:=CHR(255-((I div 256) mod 256));
    OCTab[O+2]:=CHR(255-((I div 65536) mod 256));
    OCTab[O+3]:=CHR(255-(I div 16777216));
  end;
end;

procedure WriteOpCode;
var
  I:integer;
  CTab : array of Byte;
begin
  SetLength(CTab,CodeAddress);
  for I:=1 to CodeAddress do CTab[I]:=ord(OCTab[I]);
end;

procedure OCS(S: String);
var
  i:integer;
begin
  for i:=1 to Length(S) do OCC(S[I]);
end;

const
  locNone                               = 0;
  locPushEAX                            = 1;
  locPopEAX                             = 2;
  locPopEBX                             = 3;
  locIMulEBX                            = 4;
  locXorEDXEDX                          = 5;
  locIDivEBX                            = 6;
  locPushEDX                            = 7;
  locCmpEAXEBX                          = 8;
  locMovzxEAXAL                         = 9;
  locMovDWordPtrESPEAX                  = 10;
  locJNZJNE0x03                         = 11;
  locMovDWordPtrEBXEAX                  = 12;
  locJmpDWordPtrESIOfs                  = 13;
  locCallDWordPtrESIOfs                 = 14;
  locXChgEDXESI                         = 15;
  locPopESI                             = 16;
  locMovECXImm                          = 17;
  locCLD                                = 18;
  locREPMOVSB                           = 19;
  locTestEAXEAX                         = 20;
  locNegDWordPtrESP                     = 21;
  locMovEAXDWordPtrESP                  = 22;
  locMovEBXDWordPtrFORStateCurrentValue = 23;
  locCmpDWordPtrEBXEAX                  = 24;
  locMovEAXDWordPtrFORStateDestValue    = 25;

var
  LOCV,PC:integer;

procedure OCPushEAX;
begin
  if LOCV=locPopEAX then begin
    if Code[PC]=CodeAddress then Code[PC]:=Code[PC]-1;
    if CodeAddress>0 then CodeAddress:=CodeAddress-1;
    LOCV:=locNone;
  end else begin
    OC($50);
    LOCV:=locPushEAX;
  end;
end;

procedure OCPopEAX;
begin
  if LOCV=locPushEAX then begin
    if Code[PC]=CodeAddress then Code[PC]:=Code[PC]-1;
    if CodeAddress>0 then CodeAddress:=CodeAddress-1;
    LOCV:=locNone;
  end else begin
    OC($58);
    LOCV:=locPopEAX;
  end;
end;

procedure OCPopEBX;                             begin OC($5b); LOCV:=locPopEBX;end;
procedure OCIMulEBX;                            begin OC($f7); OC($eb);LOCV:=locIMulEBX;end;
procedure OCXorEDXEDX;                          begin OC($33); OC($d2);LOCV:=locXorEDXEDX;end;
procedure OCIDIVEBX;                            begin OC($f7); OC($fb);LOCV:=locIDivEBX;end;
procedure OCPushEDX;                            begin OC($52); LOCV:=locPushEDX;end;
procedure OCCmpEAXEBX;                          begin OC($3b); OC($c3);LOCV:=locCmpEAXEBX;end;
procedure OCMovzxEAXAL;                         begin OC($0f); OC($b6); OC($c0);LOCV:=locMovzxEAXAL;end;
procedure OCJNZJNE0x03;                         begin OC($75); OC($03);LOCV:=locJNZJNE0x03;end;
procedure OCMovDWordPtrESPEAX;                  begin OC($89); OC($04); OC($24);LOCV:=locMovDWordPtrESPEAX;end;
procedure OCMovDWordPtrEBXEAX;                  begin OC($89); OC($03);LOCV:=locMovDWordPtrEBXEAX;end;
procedure OCJmpDWordPtrESIOfs(Ofs:integer);     begin OC($ff); OC($66); OC(Ofs);LOCV:=locJmpDWordPtrESIOfs;end;
procedure OCCallDWordPtrESIOfs(Ofs:integer);    begin OC($ff); OC($56); OC(Ofs);LOCV:=locCallDWordPtrESIOfs;end;
procedure OCXChgEDXESI;                         begin OC($87); OC($d6);LOCV:=locXChgEDXESI;end;
procedure OCPopESI;                             begin OC($5e); LOCV:=locPopESI;end;
procedure OCMovECXImm(Value:integer);           begin OC($b9); OCI(Value);LOCV:=locMovECXImm;end;
procedure OCCLD;                                begin OC($fc); LOCV:=locCLD;end;
procedure OCREPMOVSB;                           begin OC($f3); OC($a4);LOCV:=locREPMOVSB;end;
procedure OCTestEAXEAX;                         begin OC($85); OC($c0); { TEST EAX,EAX }LOCV:=locTestEAXEAX;end;
procedure OCNegDWordPtrESP;                     begin OC($f7); OC($1c); OC($24); { NEG DWORD PTR [ESP] }LOCV:=locNegDWordPtrESP;end;
procedure OCMovEAXDWordPtrESP;                  begin OC($8b); OC($04); OC($24); { MOV EAX,DWORD PTR [ESP] }LOCV:=locMovEAXDWordPtrESP;end;
procedure OCMovEBXDWordPtrFORStateCurrentValue; begin OC($8b); OC($5d); OC($04);LOCV:=locMovEBXDWordPtrFORStateCurrentValue;end;
procedure OCCmpDWordPtrEBXEAX;                  begin OC($39); OC($03);LOCV:=locCmpDWordPtrEBXEAX;end;
procedure OCMovEAXDWordPtrFORStateDestValue;    begin OC($8b); OC($45); OC($08);LOCV:=locMovEAXDWordPtrFORStateDestValue;end;

var
  JmpCodePutTBL : array [1..MaxCode] of integer;

procedure AssembleAndLink;
var
  JCPC,JSPC,i,d,j,a:integer;
begin

  CodeAddress:=0;
  LOCV:=locNone;
  PC:=0;
  JCPC:=0;
  JSPC:=0;

  while PC<CodePos do begin
    I:=Code[PC];
    D:=Code[PC+1];
    Code[PC]:=CodeAddress;
    case I of
        OPAdd:
              begin
                OCPopEAX;
                OC($01); OC($04); OC($24);                  //WriteLn('ADD DWORD PTR [ESP],EAX');
                LOCV:=locNone;
              end;
        OPNeg:
              begin
                OCNegDWordPtrESP;
              end;
        OPMul:
              begin
                OCPopEBX;
                OCPopEAX;
                OCIMulEBX;
                OCPushEAX;
              end;
        OPDivD:
              begin
                OCPopEBX;
                OCPopEAX;
                OCXorEDXEDX;
                OCIDIVEBX;
                OCPushEAX;
              end;
        OPRemD:
              begin
                OCPopEBX;
                OCPopEAX;
                OCXorEDXEDX;
                OCIDIVEBX;
                OCPushEDX;
              end;
        OPDiv2:
              begin
                OC($d1); OC($3c); OC($24);                  //WriteLn('SAR DWORD PTR [ESP],1');
                LOCV:=locNone;
              end;
        OPRem2:
              begin
                OCPopEAX;
                OC($8b); OC($d8);                           //WriteLn('MOV EBX,EAX');
                OC($25); OC($01); OC($00); OC($00); OC($80);//WriteLn('AND EAX,$80000001');
                OC($79); OC($05);                           //WriteLn('JNS +$05');
                OC($48);                                    //WriteLn('DEC EAX');
                OC($83); OC($c8); OC($fe);                  //WriteLn('OR EAX,BYTE -$02');
                OC($40);                                    //WriteLn('INC EAX');
                LOCV:=locNone;
                OCIMulEBX;
                OCPushEAX;
              end;
        OPEqlI:
              begin
                OCPopEBX;
                OCPopEAX;
                OCCmpEAXEBX;
                OC($0f); OC($94); OC($d0);                   //WriteLn('SETE AL');
                LOCV:=locNone;
                OCMovzxEAXAL;
                OCPushEAX;
              end;
        OPNEqI:
              begin
                OCPopEBX;
                OCPopEAX;
                OCCmpEAXEBX;
                OC($0f); OC($95); OC($d0);                   //WriteLn('SETNE AL');
                LOCV:=locNone;
                OCMovzxEAXAL;
                OCPushEAX;
              end;
        OPLssI:
              begin
                OCPopEBX;
                OCPopEAX;
                OCCmpEAXEBX;
                OC($0f); OC($9c); OC($d0);                   //WriteLn('SETL AL');
                LOCV:=locNone;
                OCMovzxEAXAL;
                OCPushEAX;
              end;
        OPLeqI:
              begin
                OCPopEBX;
                OCPopEAX;
                OCCmpEAXEBX;
                OC($0f); OC($9e); OC($d0);                   //WriteLn('SETLE AL');
                LOCV:=locNone;
                OCMovzxEAXAL;
                OCPushEAX;
              end;
        OPGtrI:
              begin
                OCPopEBX;
                OCPopEAX;
                OCCmpEAXEBX;
                OC($0f); OC($9f); OC($d0);                   //WriteLn('SETG AL');
                LOCV:=locNone;
                OCMovzxEAXAL;
                OCPushEAX;
              end;
        OPGEqi:
              begin
                OCPopEBX;
                OCPopEAX;
                OCCmpEAXEBX;
                OC($0f); OC($9d); OC($d0);                   //WriteLn('SETGE AL');
                LOCV:=locNone;
                OCMovzxEAXAL;
                OCPushEAX;
              end;
        OPDupl:
              begin
                OC($ff); OC($34); OC($24);                   //WriteLn('PUSH DWORD PTR [ESP]');
                LOCV:=locNone;
              end;
        OPSwap:
              begin
                OCPopEBX;
                OCPopEAX;
                OC($53);                                      //WriteLn('PUSH EBX');
                LOCV:=locNone;
                OCPushEAX;
              end;
        OPAndB:
              begin
                OCPopEAX;
                OCTestEAXEAX;
                OCJNZJNE0x03;
                OCMovDWordPtrESPEAX;
                LOCV:=locNone;
              end;
        OPOrB:
              begin
                OCPopEAX;
                OC($83); OC($f8); OC($01);                     //WriteLn('CMP EAX,1');
                LOCV:=locNone;
                OCJNZJNE0x03;
                OCMovDWordPtrESPEAX;
                LOCV:=locNone;
              end;
        OPLoad:
              begin
                OCPopEAX;
                OC($ff); OC($30);                              //WriteLn('PUSH DWORD PTR [EAX]');
                LOCV:=locNone;
              end;
        OPStore:
              begin
                OCPopEBX;
                OCPopEAX;
                OCMovDWordPtrEBXEAX;
              end;
        (*OPHalt:
              begin
                OC($6a); OC(0);                                   { PUSH 0}
                OC($FF); OC($15);                                 { CALL ADDR }
                i:=AddDll('Kernel32.dll','ExitProcess',true);
                JmpIatTBL[CodeAddress]:=i;
                OCI($7FFFFFFF);
                LOCV:=locNone;
                PC:=PC+1;
              end;*)
        OPHalt:  OCCallDWordPtrESIOfs(0);
        OPWrI:  OCCallDWordPtrESIOfs(8);                          { CALL [ESI+8]}
        OPWrC:  OCCallDWordPtrESIOfs(4);                          { CALL [ESI+4]}
        OPWrL:  OCCallDWordPtrESIOfs(12);                         { CALL [ESI+12]}
        OPRdI:
              begin
                OCPopEBX;
                OCCallDWordPtrESIOfs(20);                         { CALL [ESI+20]}
                OCMovDWordPtrEBXEAX;
              end;
        OPRdC:
              begin
                OCPopEBX;
                OCCallDWordPtrESIOfs(16);                         { CALL [ESI+16]}
                OCMovzxEAXAL;
                OCMovDWordPtrEBXEAX;
              end;
        OPRdL:  OCCallDWordPtrESIOfs(24);                         { CALL [ESI+24]}
        OPEOF:
              begin
                OCCallDWordPtrESIOfs(28);                         { CALL [ESI+28]}
                OCPushEAX;
              end;
        OPEOL:
              begin
                OCCallDWordPtrESIOfs(32);                         { CALL [ESI+32]}
                OCPushEAX;
              end;
        OPLdC: //  CHAR
              begin
                if (D>=-128) and (D<=127) then begin
                  OC($6a); OC(D);                             //WriteLn('PUSH ',D);
                end else begin
                  OC($68); OCI(D);                            //WriteLn('PUSH DWORD ',D);
                end;
                LOCV:=locNone;
                PC:=PC+1;
              end;
        OPLdA:
              begin
                if D=0 then begin
                  OC($8b); OC($c5);                            //WriteLn('MOV EAX,EBP');
                end else if (D>=-128) and (D<=127) then begin
                  OC($8d); OC($45); OC(D);                     //WriteLn('LEA EAX,[EBP+BYTE ',D,']');
                end else begin
                  OC($8d); OC($85); OCI(D);                    //WriteLn('LEA EAX,[EBP+DWORD ',D,']')
                end;
                LOCV:=locNone;
                OCPushEAX;
                PC:=PC+1;
              end;
        OPLdLA:
              begin
                if D=0 then begin
                  OC($8b); OC($c4);                             //WriteLn('MOV EAX,ESP');
                end else if (D>=-128) and (D<=127) then begin
                  OC($8d); OC($44); OC($24); OC(D);             //WriteLn('LEA EAX,[ESP+BYTE ',D,']');
                end else begin
                  OC($8d); OC($84); OC($24); OCI(D);            //WriteLn('LEA EAX,[ESP+DWORD ',D,']');
                end;
                LOCV:=locNone;
                OCPushEAX;
                PC:=PC+1;
              end;
        OPLdL:
              begin
                if D=0 then begin
                  OCMovEAXDWordPtrESP;
                end else if (D>=-128) and (D<=127) then begin
                  OC($8b); OC($44); OC($24); OC(D);             //WriteLn('MOV EAX,DWORD PTR [ESP+BYTE ',D,']');
                end else begin
                  OC($8b); OC($84); OC($24); OCI(D);            //WriteLn('MOV EAX,DWORD PTR [ESP+DWORD ',D,']');
                end;
                OCPushEAX;
                PC:=PC+1;
              end;
        OPLdG:
              begin
                if (D>=-128) and (D<=127) then begin
                  OC($8b); OC($45); OC(D);                      //WriteLn('MOV EAX,DWORD PTR [EBP+BYTE ',D,']');
                end else begin
                  OC($8b); OC($85); OCI(D);                     //WriteLn('MOV EAX,DWORD PTR [EBP+DWORD ',D,']');
                end;
                LOCV:=locNone;
                OCPushEAX;
                PC:=PC+1;
              end;
        OPStL: //   
              begin
                OCPopEAX;
                D:=D-4;
                if D=0 then begin
                  OC($89); OC($04); OC($24);                    //WriteLn('MOV DWORD PTR [ESP],EAX');
                end else if (D>=-128) and (D<=127) then begin
                  OC($89); OC($44); OC($24); OC(D);             //WriteLn('MOV DWORD PTR [ESP+BYTE ',D,'],EAX');
                end else begin
                  OC($89); OC($84); OC($24); OCI(D);            //WriteLn('MOV EAX,DWORD PTR [ESP+DWORD ',D,']');
                end;
                LOCV:=locNone;
                PC:=PC+1;
              end;
        OPStG: //   
              begin
                OCPopEAX;
                if (D>=-128) and (D<=127) then begin
                  OC($89); OC($45); OC(D);                      //WriteLn('MOV DWORD PTR [EBP+BYTE ',D,'],EAX');
                end else begin
                  OC($89); OC($85); OCI(D);                     //WriteLn('MOV EAX,DWORD PTR [EBP+DWORD ',D,']');
                end;
                LOCV:=locNone;
                PC:=PC+1;
              end;
        OPMove:
              begin
                OCXChgEDXESI;
                OC($5f);                                         //WriteLn('POP EDI');
                LOCV:=locNone;
                OCPopESI;
                OCMovECXImm(D);
                OCCLD;
                OCREPMOVSB;
                OCXChgEDXESI;
                PC:=PC+1;
              end;
        OPCopy:
              begin
                OCXChgEDXESI;
                OCPopESI;
                OCMovECXImm(D);
                OC($2b); OC($e1);                                 //WriteLn('SUB ESP,ECX');
                OC($8b); OC($fc);                                 //WriteLn('MOV EDI,ESP');
                LOCV:=locNone;
                OCCLD;
                OCREPMOVSB;
                OCXChgEDXESI;
                PC:=PC+1;
              end;
        OPAddC:
              begin
                if (D>=-128) and (D<=127) then begin
                  OC($83); OC($04); OC($24); OC(D);               //WriteLn('ADD DWORD PTR [ESP],BYTE ',D);
                end else begin
                  OC($81); OC($04); OC($24); OCI(D);              //WriteLn('ADD DWORD PTR [ESP],DWORD ',D)
                end;
                LOCV:=locNone;
                PC:=PC+1;
              end;
        OPMulC:
              begin
                if D=(-1) then begin
                  OCNegDWordPtrESP;
                end else if (D>=-128) and (D<=127) then begin
                  OCPopEAX;
                  OC($6b); OC($c0); OC(D);                        //WriteLn('IMUL EAX,BYTE ',D);
                  LOCV:=locNone;
                  OCPushEAX;
                end else begin
                  OCPopEAX;
                  OC($69); OC($c0); OCI(D);                       //WriteLn('IMUL EAX,DWORD ',D);
                  LOCV:=locNone;
                  OCPushEAX;
                end;
                PC:=PC+1;
              end;
        OPJmp:
              begin
                if D<>(PC+2) then begin
                  JCPC:=JCPC+1;
                  OC($e9);                                        { JMP D }
                  JmpCodePutTBL[JCPC]:=CodeAddress+1;
                  OCI(D);
                end;
                PC:=PC+1;
                LOCV:=locNone;
              end;
        OPJZ:
              begin
                JCPC:=JCPC+1;
                OCPopEAX;
                OCTestEAXEAX;
                OC($0f); OC($84);                                 { JZ D }
                JmpCodePutTBL[JCPC]:=CodeAddress+1;
                OCI(D);
                LOCV:=locNone;
                PC:=PC+1;
              end;
        OPCall:
              begin
                JCPC:=JCPC+1;
                OC($e8);                                          //WriteLn('CALL ',D);
                JmpCodePutTBL[JCPC]:=CodeAddress+1;
                OCI(D);
                LOCV:=locNone;
                PC:=PC+1;
              end;
        OPAdjS:
              begin
                if D>0 then begin
                  if (D>=-128) and (D<=127) then begin
                    OC($83); OC($c4); OC(D);                      //WriteLn('ADD ESP,BYTE ',D);
                  end else begin
                    OC($81); OC($c4); OCI(D);                     //WriteLn('ADD ESP,DWORD ',D);
                  end;
                end else if D<0 then begin
                  D:=-D;
                  if (D>=-128) and (D<=127) then begin
                    OC($83); OC($ec); OC(D);                      //WriteLn('SUB ESP,BYTE ',D);
                  end else begin
                    OC($81); OC($ec); OCI(D);                     //WriteLn('SUB ESP,DWORD ',D);
                  end;
                end;
                LOCV:=locNone;
                PC:=PC+1;
              end;
        OPExit:
              begin
                D:=D-4;
                if D>0 then begin
                  OC($c2); OCW(D);                                //WriteLn('RET ',D);
                end else if D=0 then begin
                  OC($c3);                                        //WriteLn('RET');
                end else begin
                  Error(901);
                end;
                LOCV:=locNone;
                PC:=PC+1;
              end;
    end;
    PC:=PC+1;
  end;

  { jumps + calls      }
  for i:=1 to JCPC do begin
    D:=JmpCodePutTBL[I];
    OCPI(D,((Code[OCGI(D)]-D)-3));
  end;

  WriteOpCode;
  
end;

end.
