unit Globals;

interface

uses
  SysUtils, Windows;

const
  MaxCode     = 65536;
  MaxIdent    = 512;
  MaxType     = 32;
  MaxList     = 10;
  MaxAlfa     = 20;
  MaxStrLen   = 255;
  MaxCase     = 256;
  MaxBinCode  = 262144;

  IDataRVA    = $1000;
  CodeRVA     = $2000;
  BSSRVA      = $3000;
  ImageBASE   = $400000;

  OPNone      = -1; OPAdd       = 0;  OPNeg       = 1;  OPMul       = 2;
  OPDivD      = 3;  OPRemD      = 4;  OPDiv2      = 5;  OPRem2      = 6;
  OPEqlI      = 7;  OPNEqI      = 8;  OPLssI      = 9;  OPLeqI      = 10;
  OPGtrI      = 11; OPGEqI      = 12; OPDupl      = 13; OPSwap      = 14;
  OPAndB      = 15; OPOrB       = 16; OPLoad      = 17; OPStore     = 18;
  OPHalt      = 19; OPWrI       = 20; OPWrC       = 21; OPWrL       = 22;
  OPRdI       = 23; OPRdC       = 24; OPRdL       = 25; OPEOF       = 26;
  OPEOL       = 27; OPLdC       = 28; OPLdA       = 29; OPLdLA      = 30;
  OPLdL       = 31; OPLdG       = 32; OPStL       = 33; OPStG       = 34;
  OPMove      = 35; OPCopy      = 36; OPAddC      = 37; OPMulC      = 38;
  OPJmp       = 39; OPJZ        = 40; OPCall      = 41; OPAdjS      = 42;
  OPExit      = 43;

  IdCONST     = 0;
  IdVAR       = 1;
  IdFIELD     = 2;
  IdTYPE      = 3;
  IdFUNC      = 4;

  KindSIMPLE  = 0;
  KindARRAY   = 1;
  KindRECORD  = 2;

  TypeINT     = 1;
  TypeBOOL    = 2;
  TypeCHAR    = 3;
  TypeSTR     = 4;

  FunCHR      = 0;
  FunORD      = 1;
  FunWRITE    = 2;
  FunWRITELN  = 3;
  FunREAD     = 4;
  FunREADLN   = 5;
  FunEOF      = 6;
  FunEOFLN    = 7;

type

  TIdent=record
    name          : String;
    Link          : integer;
    TypeDef       : integer;
    Kind          : integer;
    Value         : integer;
    VLevel        : integer;
    VAdr          : integer;
    RefPar        : boolean;
    Offset        : integer;
    FLevel        : integer;
    FAdr          : integer;
    LastParameter : integer;
    ReturnAddress : integer;
    Inside        : boolean;
    Extern        : boolean;
    Called        : boolean;
    ExtDll        : String;
    ExtName       : String;
  end;

  TType=record
    Size          : integer;
    Kind          : integer;
    StartIndex    : integer;
    EndIndex      : integer;
    SubType       : integer;
    Fields        : integer;
  end;

type
  EncodRec=record
       Mnem : string[8];
       RP   : SmallInt;
       Po   : byte;
  end;

var
  CurChar     : char;
  LinePos     : integer;
  LineNum     : integer;
  CurTok      : integer;
  CurID       : String;
  CurNum      : integer;
  CurReg      : byte = 0;
  CurStr      : String;
  CurStrLen   : integer;
  FuncDecl    : integer;
  LastOpcode  : integer;
  CurLevel    : integer;
  IsLabeled   : boolean;
  TokNameList : array [-1..MaxList] of integer;
  IdentPos    : integer;
  TypePos     : integer;
  IdentTab    : array [0..MaxIdent] of TIdent;
  TypeTab     : array [1..MaxType]  of TType;
  Code        : array [0..MaxCode]  of integer;
  CodePos     : integer;
  StackPos    : integer;
  ConsoleApp  : Boolean = true;
  NameFile    : String;
  Source,Dest : TextFile;

const

    TokIdent    = 0;  TokNumber   = 1;  TokStrC     = 2;  TokPlus     = 3;
    TokMinus    = 4;  TokMul      = 5;  TokLBracket = 6;  TokRBracket = 7;
    TokColon    = 8;  TokEql      = 9;  TokNEq      = 10; TokLss      = 11;
    TokLEq      = 12; TokGtr      = 13; TokGEq      = 14; TokLParent  = 15;
    TokRParent  = 16; TokComma    = 17; TokSemi     = 18; TokPeriod   = 19;
    TokAssign   = 20; TokBEGIN    = 21; TokEND      = 22; TokIF       = 23;
    TokTHEN     = 24; TokELSE     = 25; TokWHILE    = 26; TokDO       = 27;
    TokCASE     = 28; TokREPEAT   = 29; TokUNTIL    = 30; TokFOR      = 31;
    TokTO       = 32; TokDOWNTO   = 33; TokNOT      = 34; TokDIV      = 35;
    TokMOD      = 36; TokAND      = 37; TokOR       = 38; TokCONST    = 39;
    TokVAR      = 40; TokTYPE     = 41; TokARRAY    = 42; TokOF       = 43;
    TokPACKED   = 44; TokRECORD   = 45; TokPROGRAM  = 46; TokFORWARD  = 47;
    TokHALT     = 48; TokFUNC     = 49; TokPROC     = 50; TokASM      = 51;
    TokDVD      = 52;

var

  Keywords    :array [TokIdent..TokDVD] of string = (
    'IDENT',          'NUMBER',         'STRING',         '+',
    '-',              '*',              '[',              ']',
    ':',              '=',              '<>',             '<',
    '<=',             '>',              '>=',             '(',
    ')',              ',',              ';',              '.',
    ':=',             'BEGIN',          'END',            'IF',
    'THEN',           'ELSE',           'WHILE',          'DO',
    'CASE',           'REPEAT',         'UNTIL',          'FOR',
    'TO',             'DOWNTO',         'NOT',            'DIV',
    'MOD',            'AND',            'OR',             'CONST',
    'VAR',            'TYPE',           'ARRAY',          'OF',
    'PACKED',         'RECORD',         'PROGRAM',        'FORWARD',
    'HALT',           'FUNCTION',       'PROCEDURE',      'ASM',
    '/'
  );

  procedure Error(N:integer);
  function  SameText(S1,S2: String):boolean;
  function  TrimRight(const S: string): string;

implementation

//#########################   ################################

function TrimRight(const S: string): string;
var
  i: Integer;
begin
  i := Length(S);
  while (i > 0) and (S[i] <= ' ') do Dec(i);
  Result := Copy(S, 1, i);
end;

function SameText(S1,S2: String):boolean;
var
  f:boolean;
  i,c:integer;
begin
  f:=true;
  i:=1;
  c:=Length(S2);
  if c<>Length(S1) then
    f:=false
  else
  while f and (i<=c) do begin
    f:=(S1[i]=S2[i]);
    Inc(i);
  end;
  SameText:=f;
end;

function  CharToOem(s: String): String;
var
  i,code: Integer;
  ret   : String;
begin
  ret:='';
  for i:=1 to length(s) do
   begin
     code:=ord(s[i]);
     if ConsoleApp then
      case code of
        192..239: code:=code-64;
        240..255: code:=code-16;
     end else
      case code of
        128..175: code:=code+64;
        224..239: code:=code+16;
      end;
     Ret:=Ret+chr(code);
   end;
  CharToOem:=Ret;
end;

//############################   ###########################

procedure Error(N:integer);
begin
  Write(CharToOem(#10#13'    '),n:1,CharToOem(':   '),'''',NameFile,''' - ');
  case n of
    TokIdent:     Write(CharToOem('   '));
    TokNumber..TokPROC: Write('''',CharToOem(' '),CharToOem(Keywords[n]),'''');
    100:          Write(CharToOem('  '));
    101:          Write(CharToOem(' '));
    102:          Write(CharToOem('  '));
    103:          Write(CharToOem('  '));
    104:          Write(CharToOem('  '));
    105:          Write(CharToOem('Procedure is already defined'));
    106:          Write(CharToOem(' '));
    107:          Write(CharToOem('  '));
    108:          Write(CharToOem('  '));
    109:          Write(CharToOem('No such field'));
    110:          Write(CharToOem('Not an array'));
    111:          Write(CharToOem('Can not Write(CharToOem this type'));
    112:          Write(CharToOem('Can not read this type'));
    113:          Write(CharToOem('Too many arguments'));
    114:          Write(CharToOem('Passing string to var argument'));
    115:          Write(CharToOem('Passing string to not an array'));
    116:          Write(CharToOem('Passing string to array of not chars'));
    117:          Write(CharToOem('Passing string to array with mismatched size'));
    118:          Write(CharToOem('Too few arguments'));
    119:          Write(CharToOem('Using procedure in expression'));
    120:          Write(CharToOem('Using type in expression'));
    121:          Write(CharToOem('Expression expected'));
    122:          Write(CharToOem('Assigning to function'));
    123:          Write(CharToOem('Assigning to constant or type'));
    124:          Write(CharToOem('Case option should be constant'));
    125:          Write(CharToOem('Case option expected'));
    126:          Write(CharToOem('Colon expected'));
    127:          Write(CharToOem('Variable expected after FOR'));
    128:          Write(CharToOem('Iterator type is incorrect'));
    129:          Write(CharToOem('TO or DOWNTO expected'));
    130:          Write(CharToOem('Constant expected'));
    131:          Write(CharToOem('Identifier or number expected'));
    132:          Write(CharToOem('first index of array is bigger then last'));
    133:          Write(CharToOem('type expected'));
    134:          Write(CharToOem('Too many types'));
    135:          Write(CharToOem('Too many nested records'));
    136:          Write(CharToOem('Too many nested procedures'));
    137:          Write(CharToOem('Invalid return type of function'));
    138:          Write(CharToOem('More arguments then in forward declaration'));
    139:          Write(CharToOem('Argument name does not match forward declaration'));
    140:          Write(CharToOem('Argument type does not match forward declaration'));
    141:          Write(CharToOem('Argument var does not match forward declaration'));
    142:          Write(CharToOem('Less arguments then in forward declaration'));
    143:          Write(CharToOem('Already forward declared'));
    144:          Write(CharToOem('No definition for forward declared'));
    801:          Write(CharToOem('DLL name is expected'));
    802:          Write(CharToOem('Invalid DLL name'));
    803:          Write(CharToOem('External function name is expected'));
    804:          Write(CharToOem('Invalid compiler directive'));
    805:          Write(CharToOem('Unterminated comment'));
    850:          Write(CharToOem(' '));
    901:          Write(CharToOem('Negative stack size'));
    902:          Write(CharToOem('Binary is too big'));
    903:          Write(CharToOem('String is too long'));
    906:          Write(CharToOem('Too many case options'));
    907:          Write(CharToOem('Code is too long'));
    909:          Write(CharToOem('Can not compare this type'));
  end;
  Writeln(CharToOem('  '),LineNum:1,CharToOem('  '),LinePos:1);
  halt;
end;

end.
