unit Parser;

interface

uses
  SysUtils, Assembl, Globals, Scaner;

  procedure EnterTokbol(CurID:String;K,T:integer);
  procedure GenOp(Opcode,A:integer);
  procedure GenOp2(Opcode:integer);
  procedure Block(L:integer);

implementation

{    ,    . ...}

procedure EnterTokbol(CurID:String;K,T:integer);
var
  J:integer;
begin
  if IdentPos=MaxIdent then Error(103);
  IdentPos:=IdentPos+1;
  IdentTab[0].name:=CurID;
  J:=TokNameList[CurLevel];
  while UpperCase(IdentTab[J].name)<>UpperCase(CurID) do J:=IdentTab[J].Link;
  if J<>0 then begin
    if IdentTab[J].Kind<>IdFUNC then Error(104);
    if (Code[IdentTab[J].FAdr]<>OPJmp) or (Code[IdentTab[J].FAdr+1]>0) then Error(105);
    IdentTab[J].name[1]:='$';
    Code[IdentTab[J].FAdr+1]:=CodePos;
    FuncDecl:=J;
  end;
  IdentTab[IdentPos].name:=CurID;
  IdentTab[IdentPos].Link:=TokNameList[CurLevel];
  IdentTab[IdentPos].TypeDef:=T;
  IdentTab[IdentPos].Kind:=K;
  TokNameList[CurLevel]:=IdentPos;
end;

{        }

function Position:integer;
var
  I,J:integer;
begin
  IdentTab[0].name:=CurID;
  I:=CurLevel;
  repeat
    J:=TokNameList[I];
    while not SameText(IdentTab[J].name,CurID) do J:=IdentTab[J].Link;
    I:=I-1;
  until (I<-1) or (J<>0);
  if J=0 then begin
    Error(106);
  end;
  Position:=J;
end;

{   - }

procedure AddCode(ACode:integer);
begin
  if CodePos<=MaxCode then begin
    Code[CodePos]:=ACode;
    CodePos:=CodePos+1;
  end else begin
    Error(907);
  end;
end;

{            }

procedure GenOp(Opcode,A:integer);
begin
  case Opcode of
    OPDupl,OPEOF,OPEOL,OPLdC,OPLdA,OPLdLA,OPLdL,OPLdG:    StackPos:=StackPos - 4;
    OPNeg,OPDiv2,OPRem2,OPSwap,OPLoad,OPHalt,OPWrL,
    OPRdL,OpAddC,OPMulC,OPJmp,OPCall,OPExit:              begin              end;
    OPAdd,OPMul,OPDivD,OPRemD,OPEqlI,OPNEqI,OPLssI,
    OPLeqI,OPGtrI,OPGEqI,OPAndB,OPOrB,OPWrC,OPRdI,
    OPRdC,OPStL,OPStG,OPJZ:                               StackPos:=StackPos + 4;
    OPStore,OPWrI,OPMove:                                 StackPos:=StackPos + 8;
    OPCopy:                                               StackPos:=StackPos - A + 4;
    OPAdjS:                                               StackPos:=StackPos + A;
  end;
  if not ((((Opcode=OPAddC) or (Opcode=OPAdjS)) and (A=0)) or ((Opcode=OPMulC) and (A=1))) then begin
    if IsLabeled then begin
      AddCode(Opcode);
      if Opcode>=OPLdC then AddCode(A);
      IsLabeled:=false;
    end else if (LastOpcode=OPLdC) and (Opcode=OPAdd) then begin
      Code[CodePos-2]:=OPAddC;
    end else if (LastOpcode=OPLdC) and (Opcode=OPMul) then begin
      Code[CodePos-2]:=OPMulC;
    end else if (LastOpcode=OPLdC) and (Opcode=OPNeg) then begin
      Code[CodePos-1]:=-Code[CodePos-1];
      Opcode:=LastOpcode;
    end else if (LastOpcode=OPLdC) and (Code[CodePos-1]=2) and (Opcode=OPDivD) then begin
      Code[CodePos-2]:=OPDiv2;
      CodePos:=CodePos-1;
    end else if (LastOpcode=OPLdC) and (Code[CodePos-1]=2) and (Opcode=OPRemD) then begin
      Code[CodePos-2]:=OPRem2;
      CodePos:=CodePos-1;
    end else if (LastOpcode=OPLdA) and (Opcode=OPStore) then begin
      Code[CodePos-2]:=OPStG;
    end else if (LastOpcode=OPLdA) and (Opcode=OPLoad) then begin
      Code[CodePos-2]:=OPLdG;
    end else if (LastOpcode=OPLdLA) and (Opcode=OPStore) then begin
      Code[CodePos-2]:=OPStL;
    end else if (LastOpcode=OPLdLA) and (Opcode=OPLoad) then begin
      Code[CodePos-2]:=OPLdL;
    end else begin
      AddCode(Opcode);
      if Opcode>=OPLdC then AddCode(A);
    end;
    LastOpcode:=Opcode;
  end;
end;

procedure GenOp2(Opcode:integer); begin   GenOp(Opcode,0);              end;
function CodeLabel:integer; begin CodeLabel:=CodePos; IsLabeled:=true;  end;

{     }

procedure GenAddress(Level,Address:integer);
begin
  if Level=0 then begin
    GenOp(OPLdA,Address);
  end else if Level=CurLevel then begin
    GenOp(OPLdLA,Address-StackPos);
  end else begin
    GenOp(OPLdL,-StackPos);
    while Level+1<>CurLevel do begin
      GenOp2(OPLoad);
      Level:=Level+1;
    end;
    GenOp(OPAddC,Address);
  end;
end;

procedure GenAddressVar(IdentNr:integer);
begin
 GenAddress(IdentTab[IdentNr].VLevel,IdentTab[IdentNr].VAdr);
 if IdentTab[IdentNr].RefPar then GenOp2(OPLoad);
end;

{     }

procedure MustBe(X,Y:integer);
begin
  if X<>Y then
    if (TypeTab[X].Kind=KindARRAY) and (TypeTab[Y].Kind=KindARRAY)
    and (TypeTab[X].StartIndex=TypeTab[Y].StartIndex)
    and (TypeTab[X].EndIndex=TypeTab[Y].EndIndex) then
      MustBe(TypeTab[X].SubType,TypeTab[Y].SubType)
    else
      Error(107);
end;

{   }

procedure Expression(var X:integer); forward;

procedure Selector(var T,IdentNr:integer);
var
  J,X:integer;
begin
  T:=IdentTab[IdentNr].TypeDef;
  GetToken;
  if (CurTok=TokPeriod) or (CurTok=TokLBracket) then begin
    GenAddressVar(IdentNr);
    IdentNr:=0;
    while (CurTok=TokPeriod) or (CurTok=TokLBracket) do begin
      case CurTok of
        TokPeriod:
                  begin
                    if TypeTab[T].Kind<>KindRECORD then Error(108);   { .. }
                    GetToken;
                    Check(TokIdent);
                    J:=TypeTab[T].Fields;
                    IdentTab[0].name:=CurID;
                    while not SameText(IdentTab[J].name,CurID) do J:=IdentTab[J].Link;
                    if J=0 then Error(109);
                    GenOp(OPAddC,IdentTab[J].Offset);
                    T:=IdentTab[J].TypeDef;
                    GetToken;
                  end;
        TokLBracket:                                                  { [ }
                  begin
                    repeat
                      if TypeTab[T].Kind<>KindARRAY then Error(110);
                      GetToken;
                      Expression(X);
                      MustBe(TypeINT,X);
                      GenOp(OPAddC,-TypeTab[T].StartIndex);
                      T:=TypeTab[T].SubType;
                      GenOp(OPMulC,TypeTab[T].Size);
                      GenOp2(OPAdd);
                    until CurTok<>TokComma;
                    Expect(TokRBracket);
                  end;
      end;
    end;
  end;
end;

procedure VarPar(var T:integer);
var
  J:integer;
begin
  Check(TokIdent);
  J:=Position;
  Selector(T,J);
  if J<>0 then GenAddressVar(J);
end;

{    }

procedure InternalFunction(N:integer);
var
  X,code:integer;
begin
  case N of
    FunCHR:                                                           { CHR }
                begin
                  Expect(TokLParent);
                  Expression(X);
                  MustBe(TypeINT,X);
                  Expect(TokRParent)
                end;                                                  { ORD }
    FunORD:
                begin
                  Expect(TokLParent);
                  Expression(X);
                  if X<>TypeBOOL then MustBe(TypeCHAR,X);
                  Expect(TokRParent);
                end;
    FunWRITE,                                                         { WRITE }
    FunWRITELN:
                begin
                  if N=FunWRITE then Check(TokLParent);               { ( }
                  if CurTok=TokLParent then begin
                    repeat
                      GetToken;
                      if CurTok=TokStrC then begin
                        X:=1;
                        while X<=CurStrLen do begin

                          code:=ord(CurStr[X]);
                          case code of
                            192..239: code:=code-64;
                            240..255: code:=code-16;
                          end;

                          GenOp(OPLdC,code);
                          GenOp2(OPWrC);
                          X:=X+1;
                        end;
                        GetToken;
                      end else begin
                        Expression(X);
                        if CurTok=TokColon then begin
                          MustBe(TypeINT,X);
                          GetToken;
                          Expression(X);
                          MustBe(TypeINT,X);
                          GenOp2(OPWrI);
                        end else if X=TypeINT then begin
                          GenOp(OPLdC,1);
                          GenOp2(OPWrI);
                        end else if X=TypeCHAR then begin
                          GenOp2(OPWrC);
                        end else begin
                          Error(111);
                        end;
                      end;
                    until CurTok<>TokComma;
                    Expect(TokRParent)
                  end;
                  if N=FunWRITELN then GenOp2(OPWrL);
                end;
    FunREAD,                                                          { READ }
    FunREADLN:
                begin
                  if N=FunREAD then Check(TokLParent);
                  if CurTok=TokLParent then begin
                    repeat
                      GetToken;
                      VarPar(X);
                      if X=TypeINT then begin
                        GenOp2(OPRdI);
                      end else if X=TypeCHAR then begin
                        GenOp2(OPRdC);
                      end else begin
                        Error(112);
                      end;
                    until CurTok<>TokComma;
                    Expect(TokRParent);
                  end;
                  if N=FunREADLN then GenOp2(OPRdL);
                end;
    FunEOF  :   GenOp2(OPEOF);
    FunEOFLN:   GenOp2(OPEOL);
  end;
end;

{    }

procedure FunctionCall(I:integer);
var
  OldStackPos,P,X:integer;
begin
  GetToken;
  if IdentTab[I].FLevel<0 then begin
    InternalFunction(IdentTab[I].FAdr);
  end else begin
    if (IdentTab[I].TypeDef<>0) and not IdentTab[I].Extern then GenOp(OPLdC,0);
    P:=I;
    OldStackPos:=StackPos;
    if CurTok=TokLParent then begin                                  { ( }
      repeat
        GetToken;
        if P=IdentTab[I].LastParameter then Error(113);
        P:=P+1;
        if IdentTab[P].RefPar then begin
          VarPar(X);
        end else begin
          Expression(X);
          if TypeTab[X].Kind<>KindSIMPLE then GenOp(OPCopy,TypeTab[X].Size);
        end;
        if X=TypeSTR then begin
          if IdentTab[P].RefPar then Error(114);
          if TypeTab[IdentTab[P].TypeDef].Kind<>KindARRAY then Error(115);
          if TypeTab[IdentTab[P].TypeDef].SubType<>TypeCHAR then Error(116);
          if ((TypeTab[IdentTab[P].TypeDef].EndIndex-TypeTab[IdentTab[P].TypeDef].StartIndex)+1)<>CurStrLen then Error(117);
        end else begin
          MustBe(IdentTab[P].TypeDef,X);
        end;
      until CurTok<>TokComma;
      Expect(TokRParent);                                             { ) }
    end;
    if P<>IdentTab[I].LastParameter then Error(118);
    if IdentTab[I].FLevel<>0 then GenAddress(IdentTab[I].FLevel,0);
    GenOp(OPCall,IdentTab[I].FAdr);
    StackPos:=OldStackPos;
  end;
end;

{  ,     }

procedure Factor(var T:integer);
var
  i:integer;
begin
  case CurTok of
  TokIdent:
            begin
              I:=Position;
              T:=IdentTab[I].TypeDef;
              case IdentTab[I].Kind of
                  IdCONST:                                                    { CONST }
                          begin
                            GetToken;
                            GenOp(OPLdC,IdentTab[I].Value);
                          end;
                  IdVAR:                                                      { VAR }
                          begin
                            Selector(T,I);
                            if I<>0 then GenAddressVar(I);
                            if TypeTab[T].Kind=KindSIMPLE then GenOp2(OPLoad);
                          end;
                  IdFUNC: if T=0 then Error(119) else FunctionCall(I);        { FUNC }
                  IdTYPE: Error(120);                                         { TYPE }
              end;
            end;
  TokNumber:
            begin                                             { NUMBER }
              GenOp(OPLdC,CurNum);
              T:=TypeINT;
              GetToken;
            end;
  TokStrC:
            begin                                             { STR }
              I:=CurStrLen;
              while I>=1 do begin
                GenOp(OPLdC,ord(CurStr[I]));
                I:=I-1;
              end;
              T:=TypeCHAR;
              if CurStrLen<>1 then T:=TypeSTR;
              GetToken;
            end;
  TokLParent:
            begin                                             { ( }
              GetToken;
              Expression(T);
              Expect(TokRParent);
            end;
  TokNOT:
            begin                                             { NOT }
              GetToken;
              Factor(T);
              MustBe(TypeBOOL,T);
              GenOp2(OPNeg);
              GenOp(OPAddC,1);
            end;
  else
    Error(121);
  end;
end;

{    + - MOD AND }

procedure Term(var X:integer);
var
  Y:integer;
begin
  Factor(X);
  while CurTok in [TokAND,TokMul,TokDIV,TokMOD] do begin

    if CurTok=TokAND then                                             { AND }
      MustBe(TypeBOOL,X)
    else
      MustBe(TypeINT,X);

    case CurTok of
        TokMul:                                                       { * }
                begin
                  GetToken;
                  Factor(Y);
                  GenOp2(OPMul);
                end;
        TokDIV:                                                       { DIV }
                begin
                  GetToken;
                  Factor(Y);
                  GenOp2(OPDivD);
                end;                                                  { MOD }
        TokMOD:
                begin
                  GetToken;
                  Factor(Y);
                  GenOp2(OPRemD);
                end;
        TokAND:                                                       { AND }
                begin
                  GetToken;
                  Factor(Y);
                  GenOp2(OPAndB);
                end;
    end;
    MustBe(X,Y);
  end;

end;

procedure SimpleExpression(var X:integer);
var
  Y:integer;
begin

  case CurTok of
    TokPlus:
              begin                                                  { + }
                GetToken;
                Term(X);
                MustBe(TypeINT,X);
              end;
    TokMinus:
              begin                                                  { - }
                GetToken;
                Term(X);
                MustBe(TypeINT,X);
                GenOp2(OPNeg);
              end;
  else
    Term(X);
  end;

  while CurTok in [TokOR,TokPlus,TokMinus] do begin

    if CurTok=TokOR then                                             { OR }
      MustBe(TypeBOOL,X)
    else
      MustBe(TypeINT,X);

    case CurTok of
        TokPlus:                                                      { + }
                begin
                  GetToken;
                  Term(Y);
                  GenOp2(OPAdd);
                end;
        TokMinus:                                                     { - }
                begin
                  GetToken;
                  Term(Y);
                  GenOp2(OPNeg);
                  GenOp2(OPAdd);
                end;
        TokOR:                                                        { OR }
                begin
                  GetToken;
                  Term(Y);
                  GenOp2(OPOrB);
                end;
    end;
    MustBe(X,Y);
  end;

end;

{    }

procedure Expression(var X:integer);
var
  O,Y:integer;
begin
  SimpleExpression(X);
  if CurTok in [TokEql,TokNEq,TokLss,TokLEq,TokGtr,TokGEq] then begin
    if (X=TypeSTR) or (TypeTab[X].Kind<>KindSIMPLE) then Error(909);
    O:=CurTok;
    GetToken;
    SimpleExpression(Y);
    MustBe(X,Y);
    case O of                                                         {}
        TokEql: GenOp2(OPEqlI);                                       {}
        TokNEq: GenOp2(OPNEqI);                                       {}
        TokLss: GenOp2(OPLssI);                                       {}
        TokLEq: GenOp2(OPLeqI);                                       {}
        TokGtr: GenOp2(OPGtrI);                                       {}
        TokGEq: GenOp2(OPGEqI);                                       {}
    end;
    X:=TypeBOOL;
  end;
end;

{   ,   
     }

procedure Statement;
var
  L  :array [1..MaxCase] of integer;
  M,N,I,J,T,X,R,OldStackPos:integer;
begin
  case CurTok of
    TokIdent:
              begin
                I:=Position;

            { -------------------- VAR FUNC --------------------}

                case IdentTab[I].Kind of
                    IdVAR:                                          { VAR }
                          begin
                            Selector(T,I);
                            Expect(TokAssign);
                            Expression(X);
                            MustBe(T,X);
                            if I=0 then
                              GenOp2(OPSwap)
                            else
                              GenAddressVar(I);
                            if TypeTab[T].Kind=KindSIMPLE then
                              GenOp2(OPStore)
                            else
                              GenOp(OPMove,TypeTab[T].Size);
                          end;
                    IdFUNC:                                          { FUNC }
                          begin
                            if IdentTab[I].TypeDef=0 then
                              FunctionCall(I)
                            else begin
                              if not IdentTab[I].Inside then Error(122);
                              GetToken;
                              Expect(TokAssign);
                              Expression(X);
                              MustBe(IdentTab[I].TypeDef,X);
                              GenAddress(IdentTab[I].FLevel+1,IdentTab[I].ReturnAddress);
                              GenOp2(OPStore);
                            end;
                          end;
                    IdCONST,IdFIELD,IdTYPE: Error(123);
                end;
              end;

    TokASM:
              begin
                GetToken;
                Assembled;
                Expect(TokEND);
              end;
  { -------------------- IF THEN ELSE --------------------}

    TokIF:
              begin                                   { IF }
                GetToken;
                Expression(T);
                MustBe(TypeBOOL,T);
                Expect(TokTHEN);                      { THEN }
                I:=CodeLabel;
                GenOp(OPJZ,0);
                Statement;
                if CurTok=TokELSE then begin          { ELSE }
                  GetToken;
                  J:=CodeLabel;
                  GenOp(OPJmp,0);
                  Code[I+1]:=CodeLabel;
                  I:=J;
                  Statement;
                end;
                Code[I+1]:=CodeLabel;
              end;

  { -------------------- CASE --------------------}

    TokCASE:
              begin                                    { CASE }
                GetToken;
                Expression(T);
                MustBe(TypeINT,T);                     {   }
                Expect(TokOF);                         { OF }
                J:=0;
                M:=0;
                repeat
                  if J<>0 then Code[J+1]:=CodeLabel;
                  N:=M;
                  repeat
                    if N<>M then GetToken;
                    GenOp2(OPDupl);
                    if CurTok=TokIdent then begin                               // CONST
                      I:=Position;
                      if IdentTab[I].Kind<>IdCONST then Error(124);
                      GenOp(OPLdC,IdentTab[I].Value);
                    end else if CurTok=TokNumber then begin                     // NUMBER
                      GenOp(OPLdC,CurNum);
                    end else if (CurTok=TokStrC) and (CurStrLen=1) then begin   // CHAR
                      GenOp(OPLdC,ord(CurStr[1]));
                    end else begin
                     Error(125);
                    end;
                    GenOp2(OPNEqI);
                    if N=MaxCase then Error(906);
                    N:=N+1;
                    L[N]:=CodeLabel;
                    GenOp(OPJZ,0);
                    GetToken;
                  until CurTok<>TokComma;               { , }
                  if CurTok<>TokColon then Error(126);  { : }
                  J:=CodeLabel;
                  GenOp(OPJmp,0);
                  repeat
                    Code[L[N]+1]:=CodeLabel;
                    N:=N-1;
                  until N=M;
                  GetToken;
                  Statement;
                  M:=M+1;
                  L[M]:=CodeLabel;
                  GenOp(OPJmp,0);
                  if CurTok=TokSemi then GetToken;       { ; }
                until (CurTok=TokEND) or (CurTok=TokELSE);{ END, ELSE }
                Code[J+1]:=CodeLabel;
                if CurTok=TokELSE then begin             { ELSE }
                  GetToken;
                  Statement;
                  if CurTok=TokSemi then GetToken;       { ; }
                  Expect(TokEND);                        { END }
                end;
                repeat
                  Code[L[M]+1]:=CodeLabel;
                  M:=M-1;
                until M=0;
                GenOp(OPAdjS,4);
                GetToken;
              end;

  { -------------------- FOR  --------------------}

    TokFOR:
              begin                                       { FOR }
                GetToken;
                if CurTok=TokIdent then begin
                  OldStackPos:=StackPos;

                  I:=Position;
                  if IdentTab[I].Kind<>IdVAR then Error(127);
                  Selector(T,I);
                  Expect(TokAssign);
                  Expression(X);
                  MustBe(T,X);
      
                  if I=0 then GenOp2(OPSwap) else GenAddressVar(I);
                  if TypeTab[T].Kind<>KindSIMPLE then Error(128);
                  GenOp2(OPStore);

                  R:=1;
                  if CurTok=TokTO then begin              { TO }
                    Expect(TokTO);
                  end else if CurTok=TokDOWNTO then begin { DOWNTO }
                    Expect(TokDOWNTO);
                    R:=-1;
                  end else begin
                    Error(129);
                  end;

                  J:=CodeLabel;

                  if I=0 then GenOp2(OPSwap) else GenAddressVar(I);
                  GenOp2(OPLoad);
                  Expression(X);
                  MustBe(T,X);

                  if R>0 then GenOp2(OPLeqI) else GenOp2(OPGeqI);
                  N:=CodeLabel;
                  GenOp(OPJZ,0);
                  Expect(TokDO);                           { DO }

                  Statement;

                  if I=0 then GenOp2(OPSwap) else GenAddressVar(I);
                  GenOp2(OPLoad);
                  GenOp(OPAddC,R);

                  if I=0 then GenOp2(OPSwap) else GenAddressVar(I);
                  GenOp2(OPStore);
                  GenOp(OPJmp,J);
                  Code[N+1]:=CodeLabel;
                  GenOp(OPAdjS,OldStackPos-StackPos);

                end else begin
                  Expect(TokIdent);
                end;
              end;

  { -------------------- WHILE --------------------}

    TokWHILE:
              begin                               { WHILE }
                GetToken;
                I:=CodeLabel;
                Expression(T);
                MustBe(TypeBOOL,T);
                Expect(TokDO);
                J:=CodeLabel;
                GenOp(OPJZ,0);
                Statement;
                GenOp(OPJmp,I);
                Code[J+1]:=CodeLabel;
              end;

  { -------------------- REPEAT --------------------}

    TokREPEAT:
              begin                              { REPEAT }
                I:=CodeLabel;
                repeat
                  GetToken;
                  Statement;
                until CurTok<>TokSemi;
                Expect(TokUNTIL);                { UNTIL }
                Expression(T);
                MustBe(TypeBOOL,T);
                GenOp(OPJZ,I);
              end;

  { -------------------- BEGIN --------------------}

    TokBEGIN:
              begin                               { BEGIN }
                repeat
                  GetToken;
                  Statement;
                until CurTok<>TokSemi;
                Expect(TokEND);
              end;

  { -------------------- HALT --------------------}

    TokHALT:
              begin                                { HALT }
                GenOp2(OPHalt);
                GetToken;
              end;
  end;
  
end;

{   }

procedure Constant(var C,T:integer);
var
  I,S:integer;
begin
  if (CurTok=TokStrC) and (CurStrLen=1) then begin
    C:=ord(CurStr[1]);
    T:=TypeCHAR;
  end else begin

    case CurTok of
      TokPlus:
                begin
                  GetToken;
                  S:=1;
                end;
      TokMinus:
                begin
                  GetToken;
                  S:=-1;
                end;
    else
      S:=0;
    end;

    case CurTok of
      TokIdent:
                begin
                  I:=Position;
                  if IdentTab[I].Kind<>IdCONST then Error(130);
                  C:=IdentTab[I].Value;
                  T:=IdentTab[I].TypeDef;
                end;
      TokNumber:
                begin
                  C:=CurNum;
                  T:=TypeINT;
                end;
    else
      Error(131);
    end;

    if S<>0 then begin
      MustBe(T,TypeINT);
      C:=C*S;
    end;
    
  end;
  GetToken;
end;

{   }

procedure TypeDef(var T:integer); forward;

procedure ArrayType(var T:integer);
var
  X:integer;
begin
  TypeTab[T].Kind:=KindARRAY;
  GetToken;
  Constant(TypeTab[T].StartIndex,X);
  MustBe(TypeINT,X);
  Expect(TokColon);
  Constant(TypeTab[T].EndIndex,X);
  MustBe(TypeINT,X);
  if TypeTab[T].StartIndex>TypeTab[T].EndIndex then Error(132);
  if CurTok=TokComma then begin
    ArrayType(TypeTab[T].SubType);
  end else begin
    Expect(TokRBracket);
    Expect(TokOF);
    TypeDef(TypeTab[T].SubType);
  end;
  TypeTab[T].Size:=(TypeTab[T].EndIndex-TypeTab[T].StartIndex+1)*TypeTab[TypeTab[T].SubType].Size;
end;

{   }

procedure TypeDef(var T:integer);
var
  I,J,SZ,FT:integer;
begin
  if CurTok=TokPACKED then GetToken;
  if CurTok=TokIdent then begin
    I:=Position;
    if IdentTab[I].Kind<>IdTYPE then Error(133);
    T:=IdentTab[I].TypeDef;
    GetToken;
  end else begin
    if TypePos=MaxType then Error(134);
    TypePos:=TypePos+1;
    T:=TypePos;
    if CurTok=TokARRAY then begin                                    { ARRAY }
      GetToken;
      Check(TokLBracket);
      ArrayType(T);
    end else begin
      Expect(TokRECORD);                                             { RECORD }
      if CurLevel=MaxList then Error(135);
      CurLevel:=CurLevel+1;
      TokNameList[CurLevel]:=0;
      Check(TokIdent);
      SZ:=0;
      repeat
        EnterTokbol(CurID,IdFIELD,0);
        I:=IdentPos;
        GetToken;
        while CurTok=TokComma do begin
          GetToken;
          Check(TokIdent);
          EnterTokbol(CurID,IdFIELD,0);
          GetToken;
        end;
        J:=IdentPos;
        Expect(TokColon);
        TypeDef(FT);
        repeat
          IdentTab[I].TypeDef:=FT;
          IdentTab[I].Offset:=SZ;
          SZ:=SZ+TypeTab[FT].Size;
          I:=I+1;
        until I>J;
        if CurTok=TokSemi then GetToken else Check(TokEND);                                             { END }
      until CurTok<>TokIdent;
      TypeTab[T].Size:=SZ;
      TypeTab[T].Kind:=KindRECORD;
      TypeTab[T].Fields:=TokNameList[CurLevel];
      CurLevel:=CurLevel-1;
      Expect(TokEND);
    end;
  end;
end;

{  }

procedure ConstDeclaration;
var
  A:String;
  T,C:integer;
begin
  A:=CurID;
  GetToken;
  Expect(TokEql);
  Constant(C,T);
  Expect(TokSemi);
  EnterTokbol(A,IdCONST,T);
  IdentTab[IdentPos].Value:=C;
end;

procedure TypeDeclaration;
var
  A:String;
  T:integer;
begin
  A:=CurID;
  GetToken;
  Expect(TokEql);
  TypeDef(T);
  Expect(TokSemi);
  EnterTokbol(A,IdTYPE,T);
end;

procedure VarDeclaration;
var
  P,Q,T:integer;
begin
  EnterTokbol(CurID,IdVAR,0);
  P:=IdentPos;
  GetToken;
  while CurTok=TokComma do begin
    GetToken;
    Check(TokIdent);
    EnterTokbol(CurID,IdVAR,0);
    GetToken;
  end;
  Q:=IdentPos;
  Expect(TokColon);
  TypeDef(T);
  Expect(TokSemi);
  repeat
    IdentTab[P].VLevel:=CurLevel;
    StackPos:=StackPos-TypeTab[T].Size;
    IdentTab[P].TypeDef:=T;
    IdentTab[P].VAdr:=StackPos;
    IdentTab[P].RefPar:=false;
    P:=P+1;
  until P>Q;
end;

{    }

procedure NewParameter(var P,PS:integer);
var
  R:boolean;
  T:integer;
begin
  if CurTok=TokVAR then begin                  { VAR }
    R:=true;
    GetToken;
  end else begin
    R:=false;
  end;
  Check(TokIdent);                             {  }
  P:=IdentPos;
  EnterTokbol(CurID,IdVAR,0);
  GetToken;
  while CurTok=TokComma do begin               { , }
    GetToken;
    Check(TokIdent);                           {  }
    EnterTokbol(CurID,IdVAR,0);
    GetToken;
  end;
  Expect(TokColon);                            { : }
  Check(TokIdent);                             {   }
  TypeDef(T);
  while P<IdentPos do begin
    P:=P+1;
    IdentTab[P].TypeDef:=T;
    IdentTab[P].RefPar:=R;
    if R then PS:=PS+4 else PS:=PS+TypeTab[T].Size;
  end;
end;

{     }

procedure FunctionDeclaration(IstFunktion:boolean);
var
  F,P,PS,P1,P2,OldStackPos:integer;
  S: String;
begin
  GetToken;
  Check(TokIdent);
  FuncDecl:=-1;
  EnterTokbol(CurID,IdFUNC,0);
  S:=CurStr;
  GetToken;
  F:=IdentPos;
  IdentTab[F].FLevel:=CurLevel;
  IdentTab[F].FAdr:=CodeLabel;
  IdentTab[F].ExtName:=S;
  GenOp(OPJmp,0);
  if CurLevel=MaxList then Error(136);
  CurLevel:=CurLevel+1;
  TokNameList[CurLevel]:=0;
  PS:=4;
  OldStackPos:=StackPos;
  if CurTok=TokLParent then begin                                  { ( }
    repeat
      GetToken;
      NewParameter(P,PS);
    until CurTok<>TokSemi;
    Expect(TokRParent);                                             { ) }
  end;
  if CurLevel>1 then begin
    StackPos:=-4;
  end else begin
    StackPos:=0;
  end;
  IdentTab[F].ReturnAddress:=PS;
  P:=F;
  while P<IdentPos do begin
    P:=P+1;
    if IdentTab[P].RefPar then begin
      PS:=PS-4;
    end else begin
      PS:=PS-TypeTab[IdentTab[P].TypeDef].Size;
    end;
    IdentTab[P].VLevel:=CurLevel;
    IdentTab[P].VAdr:=PS;
  end;
  if IstFunktion then begin
    Expect(TokColon);                                               { : }
    Check(TokIdent);                                                { .  }
    TypeDef(IdentTab[F].TypeDef);
    if TypeTab[IdentTab[F].TypeDef].Kind<>KindSIMPLE then Error(137);
  end;
  Expect(TokSemi);
  IdentTab[F].LastParameter:=IdentPos;

  if CurTok<>TokFORWARD then begin
    if FuncDecl>=0 then begin
      P1:=FuncDecl+1;
      P2:=F+1;
      while P1<=IdentTab[FuncDecl].LastParameter do begin
        if P2>IdentTab[F].LastParameter then Error(138);
        if not SameText(IdentTab[P1].name,IdentTab[P2].name) then Error(139);
        if IdentTab[P1].TypeDef<>IdentTab[P2].TypeDef then Error(140);
        if IdentTab[P1].RefPar<>IdentTab[P2].RefPar then Error(141);
        P1:=P1+1;
        P2:=P2+1;
      end;
      if P2<=IdentTab[F].LastParameter then Error(142);
    end;
    IdentTab[F].Inside:=true;
    Block(IdentTab[F].FAdr);
    IdentTab[F].Inside:=false;
    GenOp(OPExit,IdentTab[F].ReturnAddress-StackPos);
  end else begin                                                  { FORWARD }
    if FuncDecl>=0 then Error(143);
    GetToken;
  end;
  CurLevel:=CurLevel-1;
  StackPos:=OldStackPos;                                          { ; }
  Expect(TokSemi);
end;

{  ,      }

procedure Block(L:integer);
var
  I,D,OldStackPos,OldIdentPos:integer;
begin
  OldStackPos:=StackPos;
  OldIdentPos:=IdentPos;
  while CurTok in [TokCONST,TokTYPE,TokVAR,TokFUNC,TokPROC] do begin

    case CurTok of
      TokCONST:
                begin
                  GetToken;
                  Check(TokIdent);
                  while CurTok=TokIdent do ConstDeclaration;
                end;
      TokTYPE:
                begin
                  GetToken;
                  Check(TokIdent);
                  while CurTok=TokIdent do TypeDeclaration;
                end;
      TokVAR:
                begin
                  GetToken;
                  Check(TokIdent);
                  while CurTok=TokIdent do VarDeclaration;
                end;
      TokFUNC,
      TokPROC: FunctionDeclaration(CurTok=TokFUNC);
    end;

  end;

  if L+1=CodeLabel then
    CodePos:=CodePos-1
  else
    Code[L+1]:=CodeLabel;

  if CurLevel=0 then begin
    GenOp(OPAdjS,StackPos);
  end else begin
    D:=StackPos-OldStackPos;
    StackPos:=OldStackPos;
    GenOp(OPAdjS,D);
  end;

  Statement;

  if CurLevel<>0 then GenOp(OPAdjS,OldStackPos-StackPos);
  I:=OldIdentPos+1;
  while I<=IdentPos do begin
    if IdentTab[I].Kind=IdFUNC then
      if (Code[IdentTab[I].FAdr]=OPJmp) and (Code[IdentTab[I].FAdr+1]=0) then Error(144);
    I:=I+1;
  end;
  IdentPos:=OldIdentPos;
end;

end.
